*
 * rcString.C - reference counted string class.
 *
 * Copyright (C) 1992, Christoph Streit (streit@iam.unibe.ch)
 *                     Igor Metz (metz@iam.unibe.ch)
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

include <string.h>
include <stdlib.h>
include <iostream.h>
include "rcString.h"


/___________________________________________________________ rcStringRep
/
/ reference counted rcString representation

lass rcStringRep 

  friend class rcString;

ublic:
  rcStringRep(char);
  rcStringRep(const char*);
  ~rcStringRep();
  
rivate:
  char* str;
  int len;
  int count;

rivate:
  rcStringRep(_Char_p_*);
  void ref();
  void deref();
;

nline rcStringRep::rcStringRep(char c) {
  str = new char[2]; str[0] = c; str[1] = '\0'; 
  count = len = 1;


nline rcStringRep::rcStringRep(const char* s) {
  len = strlen(s);
  str = new char[len+1];
  strcpy(str, s);
  count = 1;


nline rcStringRep::~rcStringRep() {
  delete [] str; 


nline rcStringRep::rcStringRep(_Char_p_* r) { 
  str = *r; count = 1; len = strlen(str); 


nline void rcStringRep::ref() {
  count++;


nline void rcStringRep::deref() {
  if (--count <= 0) delete this;


/___________________________________________________________ rcString

/ constructors and destructors

cString::rcString() {
  rep = new rcStringRep(""); 


cString::rcString(char c) {
  rep = new rcStringRep(c); 


cString::rcString(const char* s) { 
  rep = new rcStringRep(s); 


cString::rcString(const rcString &s) { 
  s.rep->ref(); rep = s.rep; 


cString::rcString(_Char_p_* r) { 
  rep = new rcStringRep(r); 


cString::~rcString() { 
  rep->deref(); 


/ status reports

nsigned rcString::length() const {
  return rep->len; 


nt rcString::empty() const { 
  return rep->len == 0; 


onst char* rcString::chars() const {
  return rep->str; 


cString::operator const char*() const {
  return rep->str; 


/ element extraction and substring

har& rcString::operator[](unsigned i)

  if (i>=length()) {
    cerr << "Error: rcString::operator[] index out of range" << endl;
    exit(1);
  }
  /*
   * Copy on write!
   */
  if (rep->count > 1) {
    rep->deref();
    rep = new rcStringRep(rep->str);
  }

  return rep->str[i];


har rcString::operator[](unsigned i) const

  if (i>=length()) {
    cerr << "Error: rcString::operator[] index out of range" << endl;
    exit(1);
  }

  return rep->str[i];


cString rcString::operator()(unsigned i, unsigned j)

  if (j<i || i>=length() || j>=length()) {
    cerr << "Error: rcString::operator() index out of range" << endl;
    exit(1);
  }

  char* buf = new char[j-i+1];
  strncpy(buf, rep->str+i, j-i);
  buf[j-i] = '\0';

  return rcString(&buf);


/ assignment

onst rcString& rcString::operator=(char c)

  rep->deref();
  rep = new rcStringRep(c);

  return *this;


onst rcString& rcString::operator=(const char* s)

  rep->deref();
  rep = new rcStringRep(s);

  return *this;


onst rcString& rcString::operator=(const rcString& s)

  s.rep->ref();
  rep->deref();
  rep = s.rep;

  return *this;


/ comparison operators

nt rcString::operator==(const rcString& s) const { 
  return (strcmp(rep->str, s.rep->str) == 0); 


nt rcString::operator==(const char* s) const { 
  return (strcmp(rep->str, s) == 0); 


nt rcString::operator!=(const rcString& s) const { 
  return (strcmp(rep->str, s.rep->str) != 0); 


nt rcString::operator!=(const char* s) const { 
  return (strcmp(rep->str, s) != 0); 


nt rcString::operator<(const rcString& s) const { 
  return (strcmp(rep->str, s.rep->str) < 0); 


nt rcString::operator<=(const rcString& s) const { 
  return (strcmp(rep->str, s.rep->str) <= 0); 


nt rcString::operator>(const rcString& s) const { 
  return (strcmp(rep->str, s.rep->str) > 0); 


nt rcString::operator>=(const rcString& s) const { 
  return (strcmp(rep->str, s.rep->str) >= 0); 


/ concatenation

cString operator+(const rcString& x, const char* y)

  char* buf = new char[x.length() + strlen(y) + 1];
  strcpy(buf, x.chars());
  strcat(buf, y);

  return rcString(&buf); // call the private constructor
  

cString operator+(const char* x, const rcString& y)

  char* buf = new char[strlen(x) + y.length() + 1];
  strcpy(buf, x);
  strcat(buf, y.chars());

  return rcString(&buf); // call the private constructor
  

cString operator+(const rcString& x, const rcString& y)

  char* buf = new char[x.length() + y.length() + 1];
  strcpy(buf, x.chars());
  strcat(buf, y.chars());

  return rcString(&buf); // call the private constructor
  
  
/ stream operators

stream& operator<<(ostream& os, const rcString& s)
 
  return os << s.chars();

